// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  TU Dresden                                                            ==
// ==                                                                        ==
// ==  Institut fr Wissenschaftliches Rechnen                               ==
// ==  Zellescher Weg 12-14                                                  ==
// ==  01069 Dresden                                                         ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  https://gforge.zih.tu-dresden.de/projects/amdis/                      ==
// ==                                                                        ==
// ============================================================================

/** \file Serializer.h */

#ifndef AMDIS_SERIALIZER_H
#define AMDIS_SERIALIZER_H

#include <map>
#include "boost/lexical_cast.hpp"

#include "FileWriter.h"
#include "Parameters.h"
#include "AdaptInfo.h"

namespace AMDiS {

  template<typename ProblemType>
  class Serializer : public FileWriterInterface
  {
  public:
    Serializer(ProblemType *prob) 
      : name(""), 
	problem(prob),
	tsModulo(1), 
	timestepNumber(-1)
    {
      FUNCNAME("Serializer::Serializer()");

      GET_PARAMETER(0, problem->getName() + "->output->serialization filename", &name);
      GET_PARAMETER(0, problem->getName() + "->output->write every i-th timestep", 
		    "%d", &tsModulo);
      TEST_EXIT(name != "")("no filename\n");

#if HAVE_PARALLEL_DOMAIN_AMDIS
      name += ".p" + boost::lexical_cast<std::string>(MPI::COMM_WORLD.Get_rank());
#endif
    }

    virtual ~Serializer() {}

    virtual void writeFiles(AdaptInfo *adaptInfo, 
			    bool force,
			    int level = -1,
			    Flag traverseFlag = Mesh::CALL_LEAF_EL,
			    bool (*writeElem)(ElInfo*) = NULL) 
    {
      FUNCNAME("Serializer::writeFiles()");

      timestepNumber++;
      timestepNumber %= tsModulo;
      if ((timestepNumber != 0) && !force)
	return;

      TEST_EXIT(adaptInfo)("No AdaptInfo\n");

      std::ofstream out(name.c_str());
      TEST_EXIT(out.is_open())("Cannot open serialization file!\n");
      problem->serialize(out);
      adaptInfo->serialize(out);
      out.close();

      MSG("problem serialized to %s \n", name.c_str());
    }

  protected:
    /// Name of file to which the problem is serialized.
    std::string name;

    /// Pointer to the problem.
    ProblemType *problem;

    /// The problem is serialized every tsModulo-th timestep.
    int tsModulo;

    /// Current timestep number.
    int timestepNumber;
  };

  namespace SerUtil {

    template<typename T>
    void serialize(std::ostream& out, T& data)
    {
      out.write(reinterpret_cast<const char*>(&data), sizeof(T));
    }   

    template<typename T>
    void deserialize(std::istream& in, T& data)
    {
      in.read(reinterpret_cast<char*>(&data), sizeof(T));
    }   


    template<typename T>
    void serialize(std::ostream& out, std::vector<T>& data)
    {
      int vecSize = data.size();
      serialize(out, vecSize);
      for (typename std::vector<T>::iterator it = data.begin(); 
	   it != data.end(); ++it) {
	T v = *it;
	serialize(out, v);
      }
    }

    template<typename T>
    void deserialize(std::istream& in, std::vector<T>& data)
    {
      int vecSize = 0;
      deserialize(in, vecSize);
      data.resize(vecSize);

      for (int i = 0; i < vecSize; i++) {
	T v;
	deserialize(in, v);
	data[i] = v;
      }
    }


    template<typename T1, typename T2>
    void serialize(std::ostream& out, std::map<T1, T2>& data)
    {
      int mapSize = data.size();
      serialize(out, mapSize);

      for (typename std::map<T1,T2>::iterator it = data.begin(); 
	   it != data.end(); ++it) {
	T1 v1 = it->first;
	T2 v2 = it->second;
	serialize(out, v1);
	serialize(out, v2);
      }
    }

    template<typename T1, typename T2>
    void deserialize(std::istream& in, std::map<T1, T2>& data)
    {
      int mapSize = 0;
      deserialize(in, mapSize);

      for (int i = 0; i < mapSize; i++) {
	T1 v1;
	T2 v2;
	deserialize(in, v1);
	deserialize(in, v2);
	data[v1] = v2;
      }
    }

  }

}
#endif
